<?php

// Parse WRY source to make it HTML-compatible.
// Let's talk about tables:
// CSS hates them, but they did ok for formatting.  These aren't the tables you're looking for.
// We'll have a couple to parse this:
//   SUB tables
//   Label Tables
//
// So, we'll also have some functions:
//  Add sub
//  Add label
//  Sub Body
//  Add Else
// Ok, I'm not sure how to handle the call to pretitle programmatically.  Oh well, let's just add another function (see above)

define("VERBOSEP1", false);  // Be verbose in pass 1.
define("VERBOSEP2", true);  // Be verbose in pass 2.

define("VERBOSEP1L2", false); //Level 2 verbosity.
define("VERBOSEP2L2", true);

function add($name, &$table)
  {
  $table[]=trim($name);
  }

function printtable($table)
  {
  //Spit it out!
  foreach ($table as $line)
    {
    print "<br>";
    print $line;
    }
  }

function addsub($name, $line, &$table)
  {
  $name=trim($name);
  $line=trim($line);
  $table[$name][]=$line;
  if (VERBOSEP1L2)
    {
    print "[addsub] Added " . $line . " to sub list in " . $name . " sub. <br>";
    }
  }

function printtablesub($name)
  {
  //Spit it out!
  $table = array_keys($name);
  foreach ($table as $key)
    {
    print "<br>-- SUBROUTINE $key --</br>";
    foreach ($name[$key] as $line)
      {
      print "<br>";
      print $line;
      }
    print "<hr>";
    }
  if (VERBOSEP1L2)
    {
    print_r($name);
    }
  }


function parse_declared_sub($line)
  {
  global $subtable;
  $line=preg_replace('/DECLARE SUB /', '', $line);
  $line=preg_replace('/\(\)/', '', $line);
  add($line, $subtable);
  if (VERBOSEP1)
    {
    print "<p>Declared Sub </p>";
    }
  }

function parse_label($line)
  {
  global $labeltable;
  $line=preg_replace('/\:.*/', '', $line);
  add($line, $labeltable);
  }

function parse_insub($line, $name)
  {
  //TODO: Add error check: Name can't be empty
  global $insub;

  addsub($name, $line, $insub);
  if (VERBOSEP1)
    {
    print "<br>\n" . "In sub [" . $name . "]: " . $line . "<br>\n";
    }
  }

function parse()
  {
  global $other, $line_count;
  //Read the file into a string.  SLURP!
  $program = file("wry.bas");

  $i=0;
  $in_sub=false;
  $name="";
  foreach ($program as $line)
    {
    //This may cause problems, but I don't think so
    $line=trim($line);
    if (VERBOSEP1)
      {
      $i++;
      print $i;
      print ". " . $line;
      }

    //DECLARED SUB
    if (1==preg_match('/^$/', trim($line), $matched))
      {
      //Blank line, count it for stats but don't do anything else
      global $blank_count;
      $blank_count++;
      }
    else if (1==preg_match('/DECLARE SUB .*$/', $line, $matched))
      {
      parse_declared_sub($matched[0]);
      }
    else if (1==preg_match('/SUB.*/', $line, $matched))
      {
      // Part of a subroutine
      //Strip the SUB keyword
      $matched[0]=preg_replace('/SUB /', '', $matched[0]);
      $name=$matched[0];
      $in_sub=true;
      global $sub_overhead;
      $sub_overhead++;
      }
    else if (1==preg_match('/.*\:.*\:.*/', $line))
      {
      //MULTILINE STATEMENT
      if (VERBOSEP1)
        {
        print "<p>Found multiline statement: $line</p>";
        }
      if ($in_sub)
        {
        $lines = explode(":", $line);
        foreach ($lines as $singleline)
          {
          global $multiline_count;
          $multiline_count++;
          parse_insub($singleline, $name);

          if (VERBOSEP1)
            {
            print "<p> Added statement $singleline to $name</p>";
            }
          }
        }
      }
    else if (1==preg_match('/END SUB/', $line, $matched))
      {
      $in_sub=false;
      unset ($name);
      global $sub_overhead;
      $sub_overhead++;
      }
    else if ($in_sub)
      {
      parse_insub($line, $name);
      }
    else
      {
      global $no_label;
      $no_label++;
      add($line, $other);
      }
    $line_count++;
    if (VERBOSEP1)
      {
      print "<br>\n";
      }
    }
  }

function stats_count($insub)
  {
  $numlines=0;
  //Multidimensional array needs special treatment.
  foreach ($insub as $key)
    {
    $numlines=$numlines+count($key);
    }
  return $numlines;
  }

function stats($title, $show_tables)
  {
  //Display the count of the various items.
  global $subtable, $labeltable, $other, $insub, $blank_count, $line_count, $sub_overhead, $multiline_count, $function_count, $no_label, $ignore_count;

  print "<h1>Statistics for " . $title . "</h2>";

  if ($show_tables)
    {
    print "<h1>Sub Table</h1>";
    printtable($subtable);

    print "<h1>Label Table</h1>";
    printtable($labeltable);
    }

  if (0!=count($other))
    {
    print "<h1>Uncategorized Lines</h1>";
    printtable($other);
    }

  if (VERBOSEP1L2)
    {
    printtablesub($insub);
    }
  print "<h2>";
  print "Statistics";
  print "</h2>";
  print "Found: <br>";
  print count($subtable) . " subroutines. <br>\n";
  print count($labeltable) . " labels. <br> \n";
  print count($other) . " uncategorized lines <br>\n";
  print $function_count . " function calls <br>\n";
  print stats_count($insub) . " lines in subroutines <br>\n";
  print $multiline_count . " multi-line statements <br>";
  print $blank_count . " blank lines. <br>\n";
  print $sub_overhead . " lines in subroutine headers. <br>";
  print $no_label . " lines not yet labeled.  (High values are normal in P1.) <br>";
  print $ignore_count . " statements ignored. <br>";
  print $line_count . " total lines parsed. <br>";
  }

function add_type($key, $type, $line)
  {
  // Adds an output line to the list.
  global $output;

  if (VERBOSEP2)
    {
    print "<font color=\"blue\">add_output: </font>";
    print $line;
    print " is of type ". $type . " ";
    }
  $out=$line;
  if ($type=="output" || "input"==$type)
    {
    $out=preg_replace('/PRINT/i', '', $out);
    $out=preg_replace('/input/i', '', $out);
    $out=preg_replace('/"/', '', $out);
    if (VERBOSEP2)
      {
      print "<font color=\"blue\"> and is now </font>";
      print $out;
      }
    }
  if ($type=="function")
    {
    global $function_count;
    $function_count++;
    }
  $output[$key][$type][]=$out;
  if (VERBOSEP2)
    {
    print "<br>";
    }
  }

function is_function($line)
  {
  global $subtable;
  foreach ($subtable as $sub)
    {
    if (trim($line)==$sub)
      {
      return true;
      }
    }
  return false;
  }

function can_ignore($line)
  {
  //Ignore these words
  $ignore_list[]="CLS";
  $ignore_list[]="WIDTH";
  $ignore_list[]="FOR";
  $ignore_list[]="CINT";
  $ignore_list[]="PSET";
  $ignore_list[]="LOCATE";
  $ignore_list[]="SOUND";
  $ignore_list[]="COLOR";
  $ignore_list[]="SLEEP";
  $ignore_list[]="END";
  $ignore_list[]="END SELECT";
  $ignore_list[]="SELECT";
//   $ignore_list[]="CLS";
//   $ignore_list[]="CLS";
//   $ignore_list[]="CLS";
//   $ignore_list[]="CLS";
  foreach ($ignore_list as $ignore)
    {
    $pattern = "/" . $ignore . ".*/";
    if (preg_match($pattern, $line))
      {
      return true;
      }
    }
  return false;
  }


function pass2()
  {
  global $insub;
  $table = array_keys($insub);
  foreach ($table as $key)
    {
    foreach ($insub[$key] as $line)
      {
      if (VERBOSEP2L2)
        {
        print "<h3>$line</h3>";
        }
      global $line_count;
      $line_count++;

      //PRINT
      if (1==preg_match('/^PRINT/', $line))
        {
        if (VERBOSEP2)
          {
          print "<p>[pass2] Output line: ";
          print $line;
          print "<br>";
          }
        add_type($key, "output", $line);
        }
      else if (can_ignore($line))
        {
        //IGNORABLE KEYWORD
        global $ignore_count;
        $ignore_count++;
        add_type($key, "ignore", $line);
        if (VERBOSEP2)
          {
          print "<p><p>[pass2] Ignoring line " . $line . "</p>";
          }
        }
      else if (1==preg_match('/^INPUT/', $line))
        {
        //INPUT
        if (VERBOSEP2)
          {
          print "<p>[pass2] Input/Output line: ";
          print $line;
          print "<br>";
          }
        add_type($key, "input", $line);
        //  MAYBE        add_input
        }
      else if (1==preg_match('/^[a-z]+.*\:.*/', $line, $matched))
        {
        // LABEL
        // Since the parser separated out the multistatement lines above, a colon should indicate a label.
        parse_label($matched[0]);
        if (VERBOSEP2)
          {
          print "<p>[pass2] Label: ";
          print $line;
          print "<br>";
          }
        add_type($key, "label", $line);
        }
      else if (1==preg_match('/[[:lower:]]*$/', $line, $matched))
        {
        //may be a function call.
        if (is_function($line))
          {
          if (VERBOSEP2)
            {
            print "<p>[pass2] Function: ";
            print $line;
            print "<br>";
            }
          global $function_count;
          $function_count++;
          add_type($key, "function", $line);
          }
        }
      else
        {
        if (VERBOSEP2)
          {
          print "<p>[pass2] No label for line: " . $line . "</p>";
          }
        global $no_label;
        $no_label++;
        add_type($key, "unlabeled", $line);
        }
      }
    }
  global $other;
  foreach ($other as $name => $line)
    {
    global $line_count;
    $line_count++;
    if (1==preg_match('/[a-z].*$/', $line, $matched))
      {
      //must be a function call.
      add_type("Main", "function", $line);
      global $function_count;
      $function_count++;
      unset ($other[$name]);
      }
    }
  }

function parsed_listing()
  {
  //Displays the result of the parsing.
  global $output;
  $i=0;
  $index = array_keys($output);

  foreach ($index as $out)
    {
    $index2=array_keys($output[$out]);
    foreach ($index2 as $out2)
      {
      $index3=array_keys($output[$out][$out2]);
      foreach ($index3 as $out3)
        {
        print $out . " [";
        print $out2 . "] ";
        print $output[$out][$out2][$out3];
        print "<br>";
        }
      }
    print "<br>";
    }
  }


$subtable=array();
$labeltable=array();
$other=array();
$insub=array();
$output=array();  //This will eventually hold a transformed $insub[][], but we'll worry about memory later.
$blank_count=0;
$line_count=0;
$sub_overhead=0;
$function_count=0;
$line_count=0;  //Reset these stats
$ignore_count=0;
$no_label=0;
$multiline_count=0;


parse();
if (VERBOSEP1)
  {
  stats("Pass 1", true);
  }
else
  {
  stats("Pass 1", false);
  }
$line_count=0;  //Reset these stats
$ignore_count=0;
$no_label=0;
$multiline_count=0;
//parsed_listing();

pass2();
if (VERBOSEP2)
  {
  stats("Pass 2", true);
  }
else
  {
  stats("Pass 2", false);
  }


parsed_listing();
?>
