<?php
//Makes the first pass through the file, separating out multi-line statements and functions.
//
// The job of this parser is to:
// (a) Get the subroutine names
// (b) Separate out multi-statement lines
// (c) Separate functions
//

define("VERBOSEP1L1", false);
define("VERBOSEP1L2", false);

function parse_multiline($line, $subname, $is_label=false)
  {
  //MULTISTATEMENT LINE
  $lines = explode(":", $line);
  $is_first=true;
  foreach ($lines as $singleline)
    {
    if ($is_label&&$is_first)
      {
      $singleline=$singleline . ":";
      }
    if (!empty($singleline))
      {
      //$singleline may be empty, thus a blank line, when the colon is the last thing on the line.
      // We don't need a blank line, so don't add it.
      add($singleline, $subname);
      stats_add("Expanded multistatement lines");
      if (VERBOSEP1L1)
        {
        wg("[parse_multiline]");
        wr(" Added statement ");
        wb($singleline);
        wr(" to ");
        wbk($subname);
        wd();
        }
      }
    $is_first=false;
    }
  }

function add($line, $name="unnamed")
  {
  //Adds a line to the code table
  global $code_table;
  $line=trim($line); //Remove whitespace fore and aft
  $name=trim($name);
  $code_table[$name][]=$line;
  stats_add("Line Count");
  if (VERBOSEP1L1)
    {
    wg("[add]");
    wr(" Added ");
    wb($line);
    wr(" to ");
    wbk($name);
    wd();
    }
  }

function parse_sub_declaration($line)
  {
  global $sub_table;
  $line=preg_replace('/DECLARE SUB /', '', $line);
  $line=preg_replace('/\(\)/', '', $line);
  $sub_table[] = $line;
  if (VERBOSEP1L1)
    {
    wb("Found sub ");
    wr($line);
    wb(" declaration.");
    wd();
    }
  return $line;
  }

function parse_sub_def($line)
  {
  $line=preg_replace('/SUB/', '', $line);
  return $line;
  }

function parse()
  {
  //Read the file into an array.  SLURP!
  $program = file("wry.bas");
  static $in_sub=false;
  $subname="";

  foreach ($program as $line)
    {
    stats_add("Total Lines");
    $line=trim($line); //Get rid of excess whitespace
    if (VERBOSEP1L2)
      {
      wr($line);
      wd();
      }

    if (1==preg_match('/^$/', $line))
      {
      //This is a blank line.  Count it for stats but don't do anything else.
       if (VERBOSEP1L2)
        {
        wbk("Blank Line.");
        wd();
        }
      stats_add("Blank Line");
      }
    else if (1==preg_match('/^DECLARE SUB .*$/', $line))
      {
      // This is a subroutine declaration
      $subname=parse_sub_declaration($line);
       if (VERBOSEP1L2)
        {
        wbk("Sub Declaration.");
        wd();
        }
      stats_add("Sub Declarations");
      }
    else if (1==preg_match('/^SUB.*/', $line))
      {
      // This is a sub definition
      $in_sub=true;
      $subname=parse_sub_def($line);
      if (VERBOSEP1L2)
        {
        wbk("Sub definition.");
        wd();
        }
      stats_add("Sub Definitions");
      }
    else if (1==preg_match('/^END SUB/', $line, $matched))
      {
      if (VERBOSEP1L2)
       {
       wbk("End definition.");
       wd();
       }
      $in_sub=false;
      }
    else if (1==preg_match('/\".*\"/', $line))
      {
      // Multistatement lines are recognized by their : separator,
      // But, if it's in quotation marks it's just a normal line. (Pass 1, anyway.)
      if ($in_sub)
        {
        add($line, $subname);
        }
      else
        {
        add($line);
        }

      if (VERBOSEP1L2)
        {
        wbk("Normal line.");
        wd();
        }
      }
    else if (1==preg_match('/^[[:upper:]]+.*\:.*/', $line))
      {
      // Multistatement line
      if ($in_sub)
        {
        parse_multiline($line, $subname);
        }
      else
        {
        parse_multiline($line, "Main");
        }
      if (VERBOSEP1L2)
       {
       wbk("Multistatement Line, no label.");
       wd();
       }
      stats_add("Multistatement Line");
      }
    else if (1==preg_match('/^[[:lower:]]+.*\:$/', $line))
      {
      //Just a label
      if ($in_sub)
        {
        add($line, $subname);
        }
      else
        {
        add($line);
        }
      if (VERBOSEP1L2)
        {
        wbk("Just a label");
        wd();
        }
      stats_add("Labels");
      }
    else if (1==preg_match('/^[[:lower:]]+.*\:/', $line))
      {
      // Things get weird... Apparently it's a label that's never used but there's a CLS?
      // Maybe just Erik's inexperience?
      if ($in_sub)
        {
        parse_multiline($line, $subname, true);
        }
      else
        {
        parse_multiline($line, "Main", true);
        }
      if (VERBOSEP1L2)
       {
       wbk("Multistatement Line, with label.");
       wd();
       }
      stats_add("Multistatement Line");
      }
    else if ($in_sub)
      {
      // This needs to be lower than the multistatement line block
      add($line, $subname);
      if (VERBOSEP1L2)
       {
       wbk("Line in subroutine.");
       wd();
       }
      stats_add("Lines in Subs");
      }
    else
      {
      add($line);
      if (VERBOSEP1L2)
       {
       wbk("Unlabeled Line.");
       wd();
       }
      stats_add("Unlabeled Line");
      }
    }
  }

function parsed_listing()
  {
  //Displays the result of the parsing.
  global $code_table;
  $index = array_keys($code_table);
  foreach ($index as $out)
    {
    $index2=array_keys($code_table[$out]);
    foreach ($index2 as $out2)
      {
      wbk("[");
      wg($out);
      wbk("] ");
      wb($code_table[$out][$out2]);
      wd();
      }
    wd();
    }
  }


$sub_table=array();
$code_table=array();


?>